unit directorylister;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, graphics, PrintersDlgs, Printers;

Const
  AnInch = 2.54;

Type

   { TDirectoryEntry }

   TDirectoryEntry = Class(TCollectionItem)
   private
     FAttr: Integer;
     FFileName: String;
     FTimeStamp: TDateTime;
   Public
     Function AttributeString : String;
     Property FileName : String Read FFileName Write FFileName;
     Property Attributes : Integer Read FAttr Write Fattr;
     Property TimeStamp : TDateTime Read FTimeStamp Write FTimeStamp;
   end;

   { TDirectoryEntries }

   TDirectoryEntries = Class(TCollection)
   private
     function GetE(Index : Integer): TDirectoryEntry;
     procedure SetE(Index : Integer; const AValue: TDirectoryEntry);
   Public
     Function AddEntry : TDirectoryEntry;
     Property Entries[Index : Integer] : TDirectoryEntry Read GetE Write SetE; default;
   end;

   TDirListContent = (dcLineNumber,dcName,dcTimeStamp,dcDate,dcTime,dcAttributes);

   { TDirListColumnItem }

   TDirListColumnItem = Class (TCollectionItem)
   private
     FContent: TDirListContent;
     FFormat: String;
     FTitle: String;
     FWidth: Double;
     function GetTitle: String;
   public
     Procedure Assign(Source : TPersistent); override;
     Function DisplayString(E : TDirectoryEntry) : String; virtual;
   Published
     Property Width : Double Read FWidth Write FWidth;
     Property Content : TDirListContent Read FContent Write FContent;
     Property TimeStampFormat : String Read FFormat Write FFormat;
     Property Title : String Read GetTitle Write FTitle;
   end;

   { TDirListColumns }

   TDirListColumns  = Class (TCollection)
   private
     function GetC(AIndex : Integer): TDirListColumnItem;
     procedure SetC(AIndex : Integer; const AValue: TDirListColumnItem);
   Public
     Function AddColumn(AContent : TDirListContent; AWidth : Double) : TDirlistColumnItem;
     Property Columns[AIndex : Integer] : TDirListColumnItem Read GetC Write SetC; default;
   end;

   TListOption = (loHidden,loDirectory,loSymlinks,loNumbered,
                  loDrawHeader,loSorted);
   TListOptions = set of TListOption;

   { TDirectoryLister }

   TDirectoryLister = Class(TComponent)
   private
     FBottomMargin: Double;
     FColumns: TDirListColumns;
     FDirectory: String;
     FDrawHeader: Boolean;
     FEntries: TDirectoryEntries;
     FFileNameWidth: Double;
     FFont: TFont;
     FLeftMargin: Double;
     FLineSpacing: Integer;
     FOptions: TListOptions;
     FRightMargin: Double;
     FTimeStampWidth: Double;
     FTitle: String;
     FTopMargin: Double;
     function DoPrinterSetup: Boolean;
     function GetDrawHeader: Boolean;
     function GetSorted: Boolean;
     procedure SetColumns(const AValue: TDirListColumns);
   Protected
     Procedure CreateDefaultColumns;
   Public
     Constructor Create(AOwner : TComponent); override;
     Destructor Destroy; override;
     procedure DrawPage(ACanvas : TCanvas; APageNo,ALinesPerPage,ADPI : Integer);
     Procedure ReadDirectory;
     Procedure Execute;
     Function GetLinesPerPage(ACanvas : TCanvas; ACanvasHeight, ADPI : Integer) : Integer;
     function Points(AUnits: Double; ADPI: Integer): Integer;
     Property Entries : TDirectoryEntries Read FEntries;
     Property DrawHeader : Boolean Read GetDrawHeader;
     Property Sorted : Boolean Read GetSorted;
   Published
     Property Directory : String Read FDirectory Write FDirectory;
     Property Font : TFont Read FFont Write FFont;
     Property Title : String Read FTitle Write FTitle;
     Property Options : TListOptions Read FOptions Write FOptions;
     Property TopMargin : Double Read FTopMargin Write FTopMargin;
     Property BottomMargin : Double Read FBottomMargin Write FBottomMargin;
     Property LeftMargin : Double Read FLeftMargin Write FLeftMargin;
     Property RighMargin : Double Read FRightMargin Write FRightMargin;
     Property LineSpacing : Integer Read FLineSpacing Write FLineSpacing;
     Property Columns :TDirListColumns Read FColumns Write SetColumns;
   end;


Const
  DefaultTitles : Array[TDirListContent] of String
                = ('No.','Name','Date/Time','Date','Time','Attributes');

implementation



procedure TDirectoryLister.SetColumns(const AValue: TDirListColumns);
begin
  if FColumns=AValue then exit;
  FColumns.Assign(AValue);
end;

function TDirectoryLister.GetDrawHeader: Boolean;
begin
  Result:=loDrawheader in Options;
end;

function TDirectoryLister.GetSorted: Boolean;
begin
  Result:=loSorted in Options;
end;

procedure TDirectoryLister.CreateDefaultColumns;

begin
  FColumns.AddColumn(dcLineNumber,1);
  FColumns.AddColumn(dcName,5);
  FColumns.AddColumn(dcDate,3).TimeStampFormat:='yyyy mm dd';
  FColumns.AddColumn(dcTime,2);
  FColumns.AddColumn(dcAttributes,2);
end;

constructor TDirectoryLister.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FEntries:=TDirectoryEntries.Create(TDirectoryEntry);
  FColumns:=TDirListColumns.Create(TDirListColumnItem);
  FLineSpacing:=2;
  FFont:=TFont.Create;
  FFont.Name:='Courier New';
  FFont.Size:=10;
  FFont.Color:=clBlack;
  FOptions:=[loHidden,loDirectory,loSymlinks,loNumbered,loSorted,loDrawHeader];
  // Assume CM as measure
  FTopMargin:=2;
  FLeftMargin:=2;
  FBottomMargin:=2;
  FRightMargin:=2;
  FFileNameWidth:=5;
  FTimeStampWidth:=4;
  FDrawHeader:=True;
end;

destructor TDirectoryLister.Destroy;
begin
  FreeAndNil(FEntries);
  FreeAndNil(FColumns);
  inherited Destroy;
end;

Function TDirectoryLister.Points(AUnits : Double; ADPI : Integer) : Integer;

begin
  Result:=ROund(AUnits*(ADPI/AnInch));
end;


procedure TDirectoryLister.DrawPage(ACanvas : TCanvas; APageNo,ALinesPerPage,ADPI : Integer);

var
  I,J,Min,Max,H,
  FMarginX,FCurrentX,FCurrentY : integer;
  W : Double;
  s: string;
  LineNum: integer;
  E : TDirectoryEntry;
  CC : Boolean;

begin
  CC:=(FColumns.Count=0);
  if cc then
    CreateDefaultColumns;
  Try
    Min:=(Pred(APageNo)*ALinesPerPage)+1;
    Max:=(APageNo*ALinesPerPage);
    If (Max>=FEntries.Count) then
      Max:=FEntries.Count-1;
    FCurrentY:=Points(FTopMargin,ADPI);
    FMarginX:=Points(FLeftMargin,ADPI);
    H:=ACanvas.TextHeight('X')+FLineSpacing;
    If FDrawHeader then
      begin
      FCurrentX:=FMarginX;
      For J:=0 to FColumns.Count-1 do
        begin
        S:=FColumns[J].Title;
        ACanvas.TextOut(FCurrentX,FCurrentY,S);
        W:=FColumns[J].Width;
        ACanvas.Line(FCurrentX,FCurrentY+Round(H*1.5),FCurrentX+Points(W,ADPI),FCurrentY+Round(H*1.5));
        FCurrentX:=FCurrentX+Points(W,ADPI);
        end;
      Inc(FCurrentY,H*2);
      end;
    for I:=Min to Max do
      begin
      FCurrentX:=FMarginX;
      E:=FEntries[I-1];
      For J:=0 to FColumns.Count-1 do
        begin
        S:=FColumns[J].DisplayString(E);
        ACanvas.TextOut(FCurrentX,FCurrentY,S);
        FCurrentX:=FCurrentX+Points(FColumns[J].Width,ADPI);
        end;
      Inc(FCurrentY,H);
      end;
  Finally
    If CC then
      FColumns.Clear;
  end;
end;

procedure TDirectoryLister.ReadDirectory;

Var
  Info : TSearchRec;
  A,I : Integer;
  D : String;
  E : TDirectoryEntry;
  L : TStringList;

begin
  FEntries.Clear;
  A:=0;
  If loHidden in FOptions then
    A:=A or faHidden;
  if loDirectory in FOptions then
    A:=A or faDirectory;
{$ifndef ver2_2}
  if loSymlinks in FOptions then
    A:=A or faSymlink;
{$endif ver2_2}
  If (Directory<>'') then
    D:=IncludeTrailingPathDelimiter(Directory)
  else
    D:='';
  D:=D+AllFilesMask;
  If FindFirst(D,A,Info)=0 then
    try
      Repeat
        E:=FEntries.AddEntry;
        E.FileName:=Info.Name;
        E.Attributes:=Info.Attr;
        E.TimeStamp:=FileDateToDateTime(Info.Time);
      Until FindNext(Info)<>0;
    finally
      FindClose(Info);
    end;
  if Sorted then
    begin
    L:=TStringList.Create;
    try
      For I:=FEntries.Count-1 downto 0 do
        begin
        L.AddObject(FEntries[I].FIleName,FEntries[i]);
        FEntries[i].Collection:=Nil;
        end;
      L.Sorted:=True;
      For I:=0 to L.Count-1 do
        TDirectoryEntry(L.Objects[i]).Collection:=FEntries;
    finally
      L.Free;
    end;
    end
end;

Function TDirectoryLister.DoPrinterSetup: Boolean;

Var
  D : TPrintDialog;

begin
  D:=TPrintDialog.Create(Self);
  Try
    Result:=D.Execute;
  Finally
    D.Free;
  end;
end;

procedure TDirectoryLister.Execute;

var
  P,PageCount,LinesPerPage : integer;
  LineHeight : Double;

begin
  If not DoPrinterSetup then
    Exit;
  Printer.Title := 'Directory Listing:'+Directory;
  Printer.BeginDoc;
  Printer.Canvas.Font := FFont;
  ReadDirectory;
  LinesPerPage:=GetLinesPerPage(Printer.Canvas,Printer.PageHeight,Printer.YDPI);
  PageCount := FEntries.Count div LinesPerPage;
  if FEntries.Count mod LinesPerPage <> 0 then
    Inc(PageCount);
  try
    //print each page
    for P := 1 to PageCount do
      begin
      DrawPage(Printer.Canvas,P,LinesPerPage,Printer.YDPI);
      if P<>PageCount then
        Printer.NewPage;
      end;
    Printer.EndDoc;
  except
    on E:Exception do
      begin
      Printer.Abort;
      raise;
      end;
  end;
end;

function TDirectoryLister.GetLinesPerPage(ACanvas: TCanvas; ACanvasHeight, ADPI : Integer): Integer;

Var
  H : Integer;
  DPC : Integer;

begin
  DPC:=Round(ADPI/AnInch);
  H := ACanvas.TextHeight('X') + LineSpacing;
  Result:=Round((ACanvasHeight-DPC*(FTopMargin-FBottomMargin))/H-3);
  If FDrawHeader then
    Dec(Result,2);
end;


{ TDirectoryEntries }

function TDirectoryEntries.GetE(Index : Integer): TDirectoryEntry;
begin
  Result:=TDirectoryEntry(Items[Index]);
end;

procedure TDirectoryEntries.SetE(Index : Integer; const AValue: TDirectoryEntry);

begin
  Items[Index]:=AValue;
end;

function TDirectoryEntries.AddEntry: TDirectoryEntry;
begin
  Result:=TDirectoryEntry(Add);
end;

{ TDirectoryEntry }

function TDirectoryEntry.AttributeString: String;

  Procedure AddChar(AChar : Char; AAttr : Integer);

  begin
    If ((FAttr and AAttr)=AAttr) then
      Result:=Result+AChar
    else
      Result:=Result+'-';
  end;

begin
  Result:='';
  AddChar('d',faDirectory);
  AddChar('h',fahidden);
  AddChar('l',faSymlink);
  AddChar('a',faArchive);
end;

{ TDirListColumnItem }

function TDirListColumnItem.GetTitle: String;
begin
  Result:=FTitle;
  If (Result='') Then
    result:=DefaultTitles[FContent];
end;

procedure TDirListColumnItem.Assign(Source: TPersistent);

Var
  C : TDirListColumnItem;

begin
  if (Source is TDirListColumnItem) then
    begin
    C:=TDirListColumnItem(Source);
    FWidth:=C.Width;
    FContent:=C.Content;
    FFormat:=C.FFormat;
    end
  else
    inherited Assign(Source);
end;

function TDirListColumnItem.DisplayString(E: TDirectoryEntry): String;
begin
  Case FContent of
    dcLineNumber : Result:=Format('%3d.',[E.Index+1]);
    dcName       : Result:=E.FileName;
    dcTimeStamp  : if (FFormat<>'') then
                     Result:=FormatDateTime(FFormat,E.TimeStamp)
                   else
                     Result:=FormatDateTime('cccc',E.TimeStamp);
    dcDate       : if (FFormat<>'') then
                     Result:=FormatDateTime(FFormat,E.TimeStamp)
                   else
                     Result:=DateToStr(E.TimeStamp);
    dcTime       : if (FFormat<>'') then
                     Result:=FormatDateTime(FFormat,E.TimeStamp)
                   else
                     Result:=TimeToStr(E.TimeStamp);
    dcAttributes : Result:=E.AttributeString;
  end;
end;

{ TDirListColumns }

function TDirListColumns.GetC(AIndex : Integer): TDirListColumnItem;
begin
  Result:=Items[AIndex] as TDirListColumnItem
end;

procedure TDirListColumns.SetC(AIndex : Integer;
  const AValue: TDirListColumnItem);
begin
  Items[AIndex]:=AValue;
end;

function TDirListColumns.AddColumn(AContent: TDirListContent; AWidth: Double
  ): TDirlistColumnItem;
begin
  Result:=Add as TDirlistColumnItem;
  Result.Content:=AContent;
  Result.Width:=AWidth;
end;

end.

