unit fsutils;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, unixtype;

Type

  { TMTabEntry }

  TMTabEntry = Class(TCollectionItem)
  private
    FDevice: String;
    FDumpFreq: Integer;
    FfsPassNo: Integer;
    FFSType: String;
    FlineNo: Integer;
    FMountOptions: String;
    FMountPoint: String;
    function GetAsString: String;
    procedure SetAsString(AValue: String);
  Public
    Procedure Assign(Source : TPersistent); override;
    Property AsString : String Read GetAsString Write SetAsString;
  Published
    Property LineNo : Integer Read FlineNo Write FLineNo;
    Property Device : String Read FDevice Write FDevice;
    Property MountPoint : String Read FMountPoint Write FMountPoint;
    Property FSType : String Read FFSType Write FFSType;
    Property MountOptions : String Read FMountOptions Write FMountOptions;
    Property fsFreq : Integer Read FDumpFreq Write FDumpFreq;
    Property fsPassNo : Integer Read FfsPassNo Write FfsPassNo;
  end;

  TMTab = Class(TCollection)
  private
    function GetE(AIndex : Integer): TMTabEntry;
    procedure SetE(AIndex : Integer; AValue: TMTabEntry);
  Public
    Function AddEntry(Const ALine : String) : TMTabEntry;
    Property Entries[AIndex : Integer] : TMTabEntry Read GetE Write SetE; default;
  end;

  { TDiskStatItem }

  TDiskStatItem = Class(TCollectionItem)
  private
    FDisk: String;
    FFreeSize: Int64;
    FSize: Int64;
  Public
    Procedure Assign(Source : TPersistent); override;
  Published
    Property Disk : String Read FDisk Write FDisk;
    Property Size : Int64 Read FSize Write FSize;
    Property FreeSize : Int64 Read FFreeSize Write FFreeSize;
  end;

  { TDiskStats }

  TDiskStats = Class(TCollection)
  private
    function GetS(AIndex : Integer): TDiskStatitem;
    procedure SetS(AIndex : Integer; AValue: TDiskStatitem);
  Public
    Function AddStat : TDiskStatitem;
    Function IndexOfStats(Const ADisk : String) : Integer;
    Function FindStats(Const ADisk : String) : TDiskStatitem;
    Function GetStats(Const ADisk : String) : TDiskStatitem;
    Property Stats[AIndex : Integer] : TDiskStatitem Read GetS Write SetS; default;
  end;

  { TCreateDiskStats }

  TCreateDiskStats = Class(TComponent)
  private
    FStats: TDiskStats;
    FAllowedFSTypes : TStrings;
    procedure SetStats(AValue: TDiskStats);
    procedure SetAFS(AValue: TStrings);
  Protected
    function AllowMTab(M: TMTabEntry): Boolean; virtual;
    procedure MTabToStats(M: TMTabEntry); virtual;
    function ReadMTab(MTab: TMTab): Integer;
  Public
    Constructor Create(AOwner : TComponent); override;
    Destructor Destroy; override;
    Function Execute : Integer; virtual;
    Property Stats : TDiskStats Read FStats Write SetStats;
    Property AllowedFSTypes : TStrings Read FAllowedFSTypes Write SetAFS;
  end;

  EDiskStats = Class(Exception);

Function FileSizeString (Size : Int64) : String;
Function ModeToSUIBit (C,RC : Char) : Char;
Function FilePermissionString (Const Mode : mode_t) : String;
Function FileOwnerString(UID : Integer) : String;
Function FileGroupString(GID : Integer) : String;

implementation

uses strutils,unix, baseunix, inifiles, fpmimetypes;

Function FileSizeString (Size : Int64) : String;

Const
  Suffixes  : Array[1..5] of String = ('Bytes','Kb','Mb','Gb','Tb');

Var
  I,M : Integer;
begin
  If (Size=-1) then
    Result:='Unknown'
  else
    begin
    I:=1;
    M:=0;
    While (Size>1024) and (I<5) do
      begin
      M:=(Size mod 1024) div 100;
      Size:=Size div 1024;
      Inc(i);
      end;
    if (M=0) then
      Result:=Format('%d %s',[Size,Suffixes[i]])
    else
      Result:=Format('%d.%d %s',[Size,M,Suffixes[i]]);
    end;
end;

Function ModeToSUIBit (C,RC : Char) : Char;

begin
  If C='x' then
    Result:=RC
  else
    Result:=Upcase(RC);
end;

Function  FilePermissionString (Const Mode : mode_t) : String;

Const
  PermissionBits : Array [1..9] of Integer =
      (S_IRUSR,S_IWUSR,S_IXUSR,
       S_IRGRP,S_IWGRP,S_IXGRP,
       S_IROTH,S_IWOTH,S_IXOTH);
  PermissionChars : Array[1..9] of char =
       ('r','w','x','r','w','x','r','w','x');
  S_ISUID       = $800;
  S_ISGID       = $400;
  S_ISVTX       = $200;

  SuidBits  : array[1..3] of Integer = (S_ISUID,S_ISGID,S_ISVTX);
  SuidChars : Array[1..3] of char = ('s','s','t') ;

Var
    i : longint;


begin
  Result:=StringOfChar('-',9);
  For I:=1 to 9 do
    If ((Mode and PermissionBits[i])=PermissionBits[i]) then
      Result[i]:=PermissionChars[i];
  For I:=1 to 3 do
    If ((Mode and SuidBits[i])=SuidBits[i]) then
      If Result[I*3]='x' then
        Result[i*3]:=SuidChars[i]
      else
        Result[i*3]:=UpCase(SuidChars[i]);
end;

Function FileOwnerString(UID : Integer) : String;

var
  L : TStringList;
  S,S1,S2,S3 : String;
  I : Integer;

begin
  Result:='';
  L:=TStringList.Create;
  try
    L.LoadFromFile('/etc/passwd');
    I:=0;
    While (Result='') and (I<L.Count) do
      begin
      S:=L[i];
      S1:=Strutils.ExtractWord(1,S,[':']);
      S3:=Strutils.ExtractWord(3,S,[':']);
      if (StrToInTDef(S3,-1)=UID) then
        Result:=S1;
      Inc(I);
      end;
  finally
    L.Free;
  end;
  If (Result='') then
     Result:=IntToStr(UID);
end;

Function FileGroupString(GID : Integer) : String;

var
  L : TStringList;
  S,S1,S2,S3 : String;
  I : Integer;

begin
  Result:='';
  L:=TStringList.Create;
  try
    L.LoadFromFile('/etc/group');
    I:=0;
    While (Result='') and (I<L.Count) do
      begin
      S:=L[i];
      S1:=Strutils.ExtractWord(1,S,[':']);
      S3:=Strutils.ExtractWord(3,S,[':']);
      if (StrToInTDef(S3,-1)=GID) then
        Result:=S1;
      Inc(I);
      end;
  finally
    L.Free;
  end;
  If (Result='') then
     Result:=IntToStr(GID);
end;


Const
  DefaultFSTypesCount = 6;
  DefaultFSTypes : Array[1..DefaultFSTypesCount] of string
                 = ('ext2','ext3','ext4','xfs','ntfs','vfat');

function TMTabEntry.GetAsString: String;

Const
  Tab = #9;

begin
  Result:=Device+Tab+MountPoint+Tab+FSType+Tab+MountOptions+Tab+IntToStr(fsFreq)+Tab+IntToStr(fsPassNo);
end;

procedure TMTabEntry.SetAsString(AValue: String);

Const
  WS = [#9,' '];

begin
  Device:=ExtractWord(1,AValue,WS);
  MountPoint:=ExtractWord(2,AValue,WS);
  FStype:=ExtractWord(3,AValue,WS);
  MountOptions:=ExtractWord(4,AValue,WS);
  fsFreq:=StrToIntDef(ExtractWord(5,AValue,WS),0);
  fsPassNo:=StrToIntDef(ExtractWord(6,AValue,WS),0);
end;

procedure TMTabEntry.Assign(Source: TPersistent);

Var
  E : TMTabEntry;

begin
  if (Source is TMTabEntry) then
    begin
    E:=Source as  TMTabEntry;
    FDevice:=E.Device;
    FDumpFreq:=E.FDumpFreq;
    FfsPassNo:=E.fsPassNo;
    FFSType:=E.FSType;
    FlineNo:=E.LineNo;
    FMountOptions:=E.MountOptions;
    FMountPoint:=E.MountPoint;
    end
  else
    inherited Assign(Source);
end;

function TMTab.GetE(AIndex : Integer): TMTabEntry;
begin
  REsult:=TMTabEntry(Items[AIndex]);
end;

procedure TMTab.SetE(AIndex : Integer; AValue: TMTabEntry);
begin
  Items[AIndex]:=AValue;
end;

function TMTab.AddEntry(const ALine: String): TMTabEntry;
begin
  Result:=Add as TMTabEntry;
  Result.AsString:=ALine;
end;

procedure TDiskStatItem.Assign(Source: TPersistent);

Var
  S : TDiskStatItem;

begin
  if (Source is TDiskStatItem) then
    begin
    S:=Source as TDiskStatItem;
    FSize:=S.Size;
    FFreeSize:=S.FreeSize;
    FDisk:=S.Disk;
    end
  else
    inherited Assign(Source);
end;

{ TCreateDiskStats }

procedure TCreateDiskStats.SetStats(AValue: TDiskStats);
begin
  if FStats=AValue then Exit;
  FStats.Assign(AValue);
end;

constructor TCreateDiskStats.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FStats:=TDiskStats.Create(TDiskStatItem);
  FAllowedFSTypes:=TStringList.Create;
  TStringList(FAllowedFSTypes).Sorted:=True;
  TStringList(FAllowedFSTypes).Duplicates:=dupIgnore;
end;

destructor TCreateDiskStats.Destroy;
begin
  FreeAndNil(FAllowedFSTypes);
  FreeAndNil(FStats);
  inherited Destroy;
end;

Function TCreateDiskStats.ReadMTab(MTab : TMTab) : Integer;
Var
  L : TStringList;
  M : TMTabEntry;
  I : integer;
begin
  MTab.Clear;
  L:=TStringList.Create;
  try
    L.LoadFromFile('/etc/mtab');
    For I:=0 to L.Count-1 do
      begin
      M:=MTab.AddEntry(L[I]);
      M.LineNo:=I+1;
      end;
    Result:=Mtab.Count;
  finally
    L.Free;
  end;
end;

procedure TCreateDiskStats.SetAFS(AValue: TStrings);
begin
  if FAllowedFSTypes=AValue then Exit;
  FAllowedFSTypes.Assign(AValue);
end;

function TCreateDiskStats.AllowMTab(M : TMTabEntry) : Boolean;

Var
  I : Integer;

begin
  If FAllowedFSTypes.Count=0 then
    For I:=1 to DefaultFSTypesCount do
      FAllowedFSTypes.Add(DefaultFSTypes[i]);
  Result:=FAllowedFSTypes.IndexOf(M.FSType)<>-1;
end;

Procedure  TCreateDiskStats.MTabToStats(M : TMTabEntry);

Var
  FSStat : TStatFS;
  MP : String;
  DStat : TDiskStatItem;

begin
  MP:=IncludeTrailingPathDelimiter(M.MountPoint)+'.';
  DStat:=FStats.AddStat;
  if (fpstatfs(Pchar(MP),@FSSTat)=0) then
    begin
    DStat.FreeSize:=FSSTat.bavail*FSSTat.bsize;
    DStat.Size:=FSSTat.blocks*FSSTat.bsize;
    DStat.Disk:=M.Device+' (on '+M.MountPoint+')';
    end;
end;

function TCreateDiskStats.Execute: Integer;

Var
  M : TMTabEntry;
  MT : TMTab;
  I : integer;

begin
  FStats.Clear;
  MT:=TMTab.Create(TMTabEntry);
  try
    ReadMTab(MT);
    For I:=0 to MT.Count-1 do
       If AllowMTab(MT[i]) then
         MTabToStats(MT[i]);
    Result:=FStats.Count;
  finally
    MT.Free;
  end;
end;

{ TDiskStats }

function TDiskStats.GetS(AIndex : Integer): TDiskStatitem;
begin
  Result:=TDiskStatitem(Items[AIndex]);
end;

procedure TDiskStats.SetS(AIndex : Integer; AValue: TDiskStatitem);
begin
  Items[AIndex]:=AValue;
end;

function TDiskStats.AddStat: TDiskStatitem;
begin
  Result:=Add as TDiskStatitem;
end;

function TDiskStats.IndexOfStats(const ADisk: String): Integer;

begin
  Result:=Count-1;
  While (Result>=0) and (Not SameFileName(GetS(Result).Disk,ADisk)) do
    Dec(Result);
end;

function TDiskStats.FindStats(const ADisk: String): TDiskStatitem;

Var
  I : Integer;

begin
  I:=IndexOfStats(ADisk);
  If (I=-1) then
    Result:=Nil
  else
    Result:=GetS(I);
end;

function TDiskStats.GetStats(const ADisk: String): TDiskStatitem;
begin
  Result:=FindStats(ADisk);
  If (Result=Nil) then
     Raise EDiskStats.CreateFmt('Unknown disk : "%s"',[ADisk]);
end;

end.

